/*
 * Based on Modem.java from AndFLmsg for Android platforms
 *
 * Partially based on Modem.java from Pskmail by Per Crusefalk and Rein Couperus
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.Tivar;

import android.app.Activity;
import android.content.*;
import android.graphics.Bitmap;
import android.media.*;
import android.media.MediaRecorder.*;
import android.net.ParseException;

import java.io.File;
import java.io.FileOutputStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;

import android.os.Process;
import android.provider.MediaStore;
import android.provider.MediaStore.Images;
import android.widget.Button;


public class Modem {


    public static final int RXMODEMIDLE = 0;
    public static final int RXMODEMSTARTING = 1;
    public static final int RXMODEMRUNNING = 2;
    public static final int RXMODEMPAUSED = 3;
    public static final int RXMODEMSTOPPING = 4;

    public static int modemState = RXMODEMIDLE;

    public static boolean modemThreadOn = true;

    public static int NumberOfOverruns = 0;
    private static AudioRecord audiorecorder = null;
    private static boolean RxON = false;
    private static int bufferSize = 0;
    private final static float sampleRate = 8000.0f;

    //Must match the rsid.h declaration: #define RSID_FFT_SIZE		1024
    public static final int RSID_FFT_SIZE = 1024;
    //Must match the rsid.h declaration: #define RSID_SAMPLE_RATE 11025.0
    public final static double RSID_SAMPLE_RATE = 11025.0f;

    private static boolean FirstBracketReceived = false;
    public static StringBuilder MonitorString = new StringBuilder();
    public static boolean WrapLF = false;
    private static StringBuilder BroadcastLine = new StringBuilder();
    private static long lastLogWriteTime = 0;
    //	private static SampleRateConversion myResampler;

    public static double frequency = 1500.0;
    public static double rsidFrequency = 0;
    static double squelch = 20.0;
    static double metric = 50; //midrange

    //Waterfall interaction
    public static double[] WaterfallAmpl = new double[RSID_FFT_SIZE];
    //Semaphore for waterfall (new array of amplitudes needed for display)
    //Is now also accessed from the c++ native side in rsid.cxx
    public static boolean newAmplReady = false;

    public static boolean stopTX = false;

    //The following value MUST Match the MAXMODES in modem.h
    public static final int MAXMODES = 300;
    //List of modems and modes returned from the C++ modems
    public static int[] modemCapListInt = new int[MAXMODES];
    public static String[] modemCapListString = new String[MAXMODES];
    public static int numModes = 0;
    //Custom list of modes as slected in the preferences (can include all modes above if 
    //  "Custom List" is not selected, or alternatively all modes manually selected in preferences)
    public static int[] customModeListInt = new int[MAXMODES];
    public static String[] customModeListString = new String[MAXMODES];
    public static int customNumModes = 0;

    //RSID Flags
    public static boolean rxRsidOn;
    public static boolean txRsidOn;

    //JD debug
    private static boolean CModemInitialized = false;

    //Mfsk picture
    public static int picBuffer[];
    private static int pixelNumber;
    public static Bitmap picBitmap = null;
    public static int mfskPicHeight = 0;
    public static int mfskPicWidth = 0;
    public static String lastSavedPictureFn = "";
    public static int slantValue = 0;


    //Declaration of native classes
    private native static String createCModem(int modemCode);

    private native static void changeCModem(int modemCode, double newFrequency);

    private native static String initCModem(double frequency);

    private native static String rxCProcess(short[] buffer, int length);

    private native static int setSquelchLevel(double squelchLevel);

    private native static double getMetric();

    private native static double getCurrentFrequency();

    private native static int getCurrentMode();

    private native static String RsidCModemReceive(float[] myfbuffer, int length, boolean doSearch);

    private native static String createRsidModem();

    private native static int[] getModemCapListInt();

    private native static String[] getModemCapListString();

    private native static void saveEnv();

    private native static void setSlowCpuFlag(boolean slowcpu);

    static {
        //Load the C++ modems library
        System.loadLibrary("c++_shared");
        System.loadLibrary("Tivar_Fldigi_Interface");
    }


    //Called from the C++ native side in rsid.cxx
    public static void updateWaterfall(double[] aFFTAmpl) {
        System.arraycopy(aFFTAmpl, 0, WaterfallAmpl, 0, RSID_FFT_SIZE);
        Modem.newAmplReady = true;
        Tivar.mHandler.post(Tivar.updatewaterfall);
    }


    //Called from the C++ native side to initialize a new picture and associated viewer. 
    public static void showRxViewer(final int mpicW, final int mpicH) {
        //Save size for later
        mfskPicWidth = mpicW;
        mfskPicHeight = mpicH;
        //Create buffer
        picBuffer = new int[mpicH * mpicW];
        //Create bitmap
        picBitmap = Bitmap.createBitmap(picBuffer, mfskPicWidth, mfskPicHeight, Bitmap.Config.ARGB_8888);
        //Reset pixel pointer
        pixelNumber = 0;
        //Open the popup window
        ((Activity) Tivar.myContext).runOnUiThread(new Runnable() {
            public void run() {
                // Display the RX picture popup window
                ((Tivar) Tivar.myContext).rxPicturePopup(mpicW, mpicH);
            }
        });
    }


    //Called from the C++ native side to update pixels in a picture. 
    public static void updatePicture(int data[], int picWidth) {
        if (picBuffer != null) {
            //Add a row to the bitmap
            int ii = 0;
            for (int i = 0; i < picWidth; i++) {
                ii = i * 3; //3 bytes per pixel
                picBuffer[pixelNumber++] = 0xff000000 | (data[ii] << 16) | (data[ii + 1] << 8) | data[ii + 2];
            }
            //Create bitmap
            picBitmap = Bitmap.createBitmap(picBuffer, mfskPicWidth, mfskPicHeight, Bitmap.Config.ARGB_8888);
            //update bitmap instead of re-creating at every line update (causes a lot of garbage collections and stops the processes)
            //Fix immutable state of Bitmap first (See Stack Overflow for backward compatible solutions)
            //try {
            //    picBitmap.setPixels(picBuffer, pixelNumber - picWidth, picWidth, 1, (pixelNumber - picWidth) / mfskPicHeight, picWidth, 1);
            //} catch (Exception e) {
            //    e.printStackTrace();
            //}
            //Display updated bitmap
            Tivar.mHandler.post(Tivar.updateMfskPicture);
        }
    }


    //Called from the C++ native side to save the last received picture. 
    public static void saveLastPicture() {

        saveAnalogPicture(true); //New picture

        //Make the de-salanting buttons visible and pre-set the de-slanting amount
        slantValue = 0;
        //Enable the edit buttons
        ((Activity) Tivar.myContext).runOnUiThread(new Runnable() {
            public void run() {
                Button editButton;
                editButton = (Button) Tivar.pwLayout.findViewById(R.id.button_slant_left);
                if (editButton != null) editButton.setEnabled(true);
                editButton = (Button) Tivar.pwLayout.findViewById(R.id.button_slant_right);
                if (editButton != null) editButton.setEnabled(true);
                editButton = (Button) Tivar.pwLayout.findViewById(R.id.button_save_again);
                if (editButton != null) editButton.setEnabled(true);
            }
        });
    }


    //Save and re-save analog picture
    public static void saveAnalogPicture(boolean newPicture) {
        String fileName = "";
        String filePath = "";

        try {
            //In any case, save the image file for further usage if required
            if (newPicture) {
                final String DATEFORMAT = "yyyy-MM-dd";
                final String TIMEFORMAT = "HHmmss";
                SimpleDateFormat sdf = new SimpleDateFormat(DATEFORMAT);
                sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
                String StrDate = sdf.format(new Date());
                SimpleDateFormat sdfTime = new SimpleDateFormat(TIMEFORMAT);
                sdfTime.setTimeZone(TimeZone.getTimeZone("UTC"));
                String StrTime = sdfTime.format(new Date());
                fileName = "pic_" + StrDate + "_" + StrTime + "z.png";
                //fileName = System.currentTimeMillis () + ".png";
                lastSavedPictureFn = fileName;
            } else {
                fileName = lastSavedPictureFn;
            }
            filePath = Processor.HomePath + Processor.Dirprefix + Processor.DirImages +
                    Processor.Separator;
            File dest = new File(filePath + fileName);
            FileOutputStream out = new FileOutputStream(dest);
            picBitmap.compress(Bitmap.CompressFormat.PNG, 100, out);
            out.flush();
            out.close();
            //Make it available in the System Picture Gallery
            ContentValues values = new ContentValues();
            values.put(Images.Media.TITLE, fileName);
            values.put(Images.Media.DATE_TAKEN, System.currentTimeMillis());
            values.put(Images.ImageColumns.BUCKET_ID, dest.toString().toLowerCase(Locale.US).hashCode());
            values.put(Images.ImageColumns.BUCKET_DISPLAY_NAME, dest.getName().toLowerCase(Locale.US));
            values.put("_data", dest.getAbsolutePath());
            values.put(Images.Media.DESCRIPTION, "Fldigi Image");
            //Display a message
            Tivar.topToastText("Saved Image: " + fileName + "\n\nUse Gallery to view and share.");
            ContentResolver cr = Tivar.myContext.getContentResolver();
            cr.insert(MediaStore.Images.Media.EXTERNAL_CONTENT_URI, values);
        } catch (Exception e) {
            loggingclass.writelog("Exception Error in 'saveLastPicture' " + e.getMessage(), null, true);
        }
    }


    //Corrects the slant in a received image. Applies the cummulative step parameter.
    public static void deSlant(int step) {

        slantValue += step;
        if (slantValue > mfskPicWidth) slantValue = mfskPicWidth;
        if (slantValue < -1 * mfskPicWidth) slantValue = -1 * mfskPicWidth;
        int pixelNumberAtLineY, sourcePixelNumber, sourcePixel, sourceNextPixel;
        int nextPixelStep = slantValue < 0 ? -1 : 1;
        int pixelShift;
        double thisPixel, nextPixel;
        double slantLineInc = (double) slantValue / (double) mfskPicHeight;
        int[] deSlantedPic = new int[picBuffer.length];
        //copy first line as-is (it is the time/spacial reference)
        for (int x = 0; x < mfskPicWidth; x++) {
            deSlantedPic[x] = picBuffer[x];
        }
        //Perform "de-slanting" from the 2n line onwards
        for (int y = 1; y < mfskPicHeight; y++) {
            pixelNumberAtLineY = mfskPicWidth * y;
            pixelShift = (int) (y * slantLineInc);
            nextPixel = Math.abs((y * slantLineInc) - (double) pixelShift);
            thisPixel = 1 - nextPixel;

            //Shift pixels using fractional value
            for (int x = 0; x < mfskPicWidth; x++) {
                if (x + pixelShift >= 0 && x + pixelShift < mfskPicWidth) {
                    sourcePixelNumber = pixelShift + pixelNumberAtLineY + x > picBuffer.length - 2 ?
                            picBuffer.length - 2 : pixelShift + pixelNumberAtLineY + x;
                    sourcePixel = picBuffer[sourcePixelNumber];
                    sourceNextPixel = picBuffer[sourcePixelNumber + nextPixelStep];
                    double red1 = (((sourcePixel & 0x00ff0000) >> 16) * thisPixel);
                    double red2 = (((sourceNextPixel & 0x00ff0000) >> 16) * nextPixel);
                    int red = (int) (red1 + red2);
                    double green1 = (((sourcePixel & 0x0000ff00) >> 8) * thisPixel);
                    double green2 = (((sourceNextPixel & 0x0000ff00) >> 8) * nextPixel);
                    int green = (int) (green1 + green2);
                    double blue1 = ((sourcePixel & 0x000000ff) * thisPixel);
                    double blue2 = ((sourceNextPixel & 0x000000ff) * nextPixel);
                    int blue = (int) (blue1 + blue2);
                    deSlantedPic[pixelNumberAtLineY + x] = 0xff000000 | (red << 16) | (green << 8) | blue;
                } else {
                    //The line is wrapped around, so need to use the previous/next line pixels
                    sourcePixelNumber = pixelShift + pixelNumberAtLineY + x - (mfskPicWidth * nextPixelStep) > picBuffer.length - 2 ?
                            picBuffer.length - 2 : pixelShift + pixelNumberAtLineY + x - (mfskPicWidth * nextPixelStep);
                    sourcePixel = picBuffer[sourcePixelNumber];
                    sourceNextPixel = picBuffer[sourcePixelNumber + nextPixelStep];
                    double red1 = (((sourcePixel & 0x00ff0000) >> 16) * thisPixel);
                    double red2 = (((sourceNextPixel & 0x00ff0000) >> 16) * nextPixel);
                    int red = (int) (red1 + red2);
                    double green1 = (((sourcePixel & 0x0000ff00) >> 8) * thisPixel);
                    double green2 = (((sourceNextPixel & 0x0000ff00) >> 8) * nextPixel);
                    int green = (int) (green1 + green2);
                    double blue1 = ((sourcePixel & 0x000000ff) * thisPixel);
                    double blue2 = ((sourceNextPixel & 0x000000ff) * nextPixel);
                    int blue = (int) (blue1 + blue2);
                    //Colors have been shifted as we transmit one line of Red, one of Green, then one of Blue
                    if (nextPixelStep > 0) { //Slant towards left
                        deSlantedPic[pixelNumberAtLineY + x] = 0xff000000 | red | (green << 16) | (blue << 8);
                    } else { //Slant towards right
                        deSlantedPic[pixelNumberAtLineY + x] = 0xff000000 | (red << 8) | green | (blue << 16);
                    }
                }
            }
        }
        picBitmap = Bitmap.createBitmap(deSlantedPic, mfskPicWidth, mfskPicHeight, Bitmap.Config.ARGB_8888);
        //Keep so that we can call the window again deSlantedPic = null; //Release for GC
        Tivar.mHandler.post(Tivar.updateMfskPicture);

    }


    //Get capability list of modems from all the C++ modems (taken from rsid_defs.cxx)
    public static void updateModemCapabilityList() {
        //get modem list (int and string description). The C++ side returns its list of available modems.
        modemCapListInt = getModemCapListInt();
        modemCapListString = getModemCapListString();
        //Now find the end of modem list to know how many different modems codes we have
        Modem.numModes = MAXMODES; //Just in case
        for (int i = 0; i < MAXMODES; i++) {
            if (modemCapListInt[i] == -1) {
                Modem.numModes = i;
                //Exit loop
                i = MAXMODES;
            }
        }
        //Sort by mode code to re-group modes of the same modem (as they are in two arrays in rsid_def.cxx)
        boolean swapped = true;
        int tmp;
        String tmpS;
        while (swapped) {
            swapped = false;
            for (int i = 0; i < numModes - 1; i++) {
                if (modemCapListInt[i] > modemCapListInt[i + 1]) {
                    tmp = modemCapListInt[i];
                    tmpS = modemCapListString[i];
                    modemCapListInt[i] = modemCapListInt[i + 1];
                    modemCapListString[i] = modemCapListString[i + 1];
                    modemCapListInt[i + 1] = tmp;
                    modemCapListString[i + 1] = tmpS;
                    swapped = true;
                }
            }
        }
        //Do we want to use a custom list of modes?
        boolean useModeList = config.getPreferenceB("USEMODELIST", false);
        if (useModeList) {
            customNumModes = 0;
            boolean thatmode;
            for (int i = 0; i < numModes; i++) {
                thatmode = config.getPreferenceB("USE" + modemCapListString[i], false);
                if (thatmode) {
                    customModeListInt[customNumModes] = modemCapListInt[i];
                    customModeListString[customNumModes++] = modemCapListString[i];
                }
            }
        } else {
            customModeListInt = modemCapListInt;
            customModeListString = modemCapListString;
            customNumModes = numModes;
        }
        //Check that we have at least one mode (if not, take the first in the capability list)
        if (customNumModes < 1) {
            customNumModes = 1;
            customModeListInt[0] = modemCapListInt[0];
            customModeListString[0] = modemCapListString[0];
        }
    }


    //Initialise RX modem
    public static void ModemInit() {
        //(re)get list of available modems
        updateModemCapabilityList();

        //Android To-DO: Change to C++ resampler instead of Java quadratic resampler
        //Initialize Re-sampling to 11025Hz for RSID, THOR and MFSK modems
        //		myResampler = new SampleRateConversion(11025.0 / 8000.0);
        SampleRateConversion.SampleRateConversionInit((float) (11025.0 / 8000.0));
        //
        BroadcastLine = new StringBuilder();
    }


    //Returns the modem code given a modem name (String)
    public static int getMode(String mstring) {
        int j = 0;
        for (int i = 0; i < modemCapListInt.length; i++) {
            if (modemCapListString[i].equals(mstring)) j = i;
        }
        return modemCapListInt[j];
    }

    //Returns the modem index in the array of modes available given a modem code
    public static int getModeIndex(int mcode) {
        int j = -1;
        for (int i = 0; i < customNumModes; i++) {
            if (mcode == customModeListInt[i]) j = i;
        }
        //In case we didn't find it, return the first mode in the list to avoid segment fault
        if (j == -1) {
            j = 0;
        }
        return j;
    }


    //Receives a modem code. Returns the modem index in the array of ALL modes supplied by the C++ modem
    public static int getModeIndexFullList(int mcode) {
        int j = -1;
        for (int i = 0; i < numModes; i++) {
            if (mcode == modemCapListInt[i]) j = i;
        }
        //In case we didn't find it, return the first mode in the list to avoid segment fault
        if (j == -1) {
            j = 0;
        }
        return j;
    }


    //Return the new mode code or the same if it hits the end of list either way
    public static int getModeUpDown(int currentMode, int increment) {
        //Find position of current mode
        int j = getModeIndex(currentMode);
        j += increment;
        //Circular list
        if (j < 0) j = customNumModes - 1;
        if (j >= customNumModes) j = 0;
        return customModeListInt[j];
    }


    private static void soundInInit() {
        bufferSize = (int) sampleRate; // 1 second of Audio max
        if (bufferSize < AudioRecord.getMinBufferSize((int) sampleRate, AudioFormat.CHANNEL_IN_MONO, AudioFormat.ENCODING_PCM_16BIT)) {
            // Check to make sure buffer size is not smaller than the smallest allowed one
            bufferSize = AudioRecord.getMinBufferSize((int) sampleRate, AudioFormat.CHANNEL_IN_MONO, AudioFormat.ENCODING_PCM_16BIT);
        }
        int ii = 20; //number of 1/4 seconds wait
        while (--ii > 0) {
            audiorecorder = new AudioRecord(AudioSource.MIC, 8000, android.media.AudioFormat.CHANNEL_IN_MONO,
                    android.media.AudioFormat.ENCODING_PCM_16BIT, bufferSize);
            if (audiorecorder.getState() == AudioRecord.STATE_INITIALIZED) {
                ii = 0;//ok done
            } else {
                if (ii < 16) { //Only if have to wait more than 1 seconds
                    loggingclass.writelog("Waiting for Audio MIC availability...", null, true);
                }
                try {
                    Thread.sleep(250);//1/4 second
                } catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        }
        if (audiorecorder.getState() != AudioRecord.STATE_INITIALIZED) {
            //Android add exception catch here
            loggingclass.writelog("Can't open Audio MIC \n", null, true);
        }
    }


    public static void startmodem() {
        modemThreadOn = true;

        new Thread(new Runnable() {
            public void run() {
                //Brings back thread priority to foreground (balance between GUI and modem)
                Process.setThreadPriority(Process.THREAD_PRIORITY_FOREGROUND);

                while (modemThreadOn) {

                    //Save Environment for this thread so it can call back
                    // Java methods while in C++
                    Modem.saveEnv();

                    String rsidReturnedString;
                    String modemReturnedString;
                    modemState = RXMODEMSTARTING;
                    double startproctime = 0;
                    double endproctime = 0;
                    int numSamples8K = 0;
                    Modem.soundInInit();
                    //debugging only
                    //Message.addEntryToLog(Message.dateTimeStamp() + "Done 'soundInInit'");
                    NumberOfOverruns = 0;
                    audiorecorder.startRecording();
                    RxON = true;
                    Processor.restartRxModem.drainPermits();
                    //Since the callback is not working, implement a while loop.
                    short[] so8K = new short[bufferSize];
                    int size12Kbuf = (int) ((bufferSize + 1) * 11025.0 / 8000.0);
                    //Android changed to float to match rsid.cxx code
                    //double[] so12K = new double[size12Kbuf];
                    float[] so12K = new float[size12Kbuf];
                    //Initialise modem
                    //debugging only
                    //Message.addEntryToLog(Message.dateTimeStamp() + "About to do 'createCModem' with Modem" + Processor.RxModem);
                    String modemCreateResult = createCModem(Processor.RxModem);
                    //debugging only
                    //Message.addEntryToLog(Message.dateTimeStamp() + "Done 'createCModem'");
                    if (modemCreateResult.contains("ERROR")) {
                        CModemInitialized = false;
                    } else {
                        CModemInitialized = true;
                        //Initialize RX side of modem
                        boolean slowCpu = config.getPreferenceB("SLOWCPU", false);
                        setSlowCpuFlag(slowCpu);
                        //Changed to getPreferencesI in case it is not an interger representation
                        //String frequencySTR = config.getPreferenceS("AFREQUENCY","1500");
                        //double centerfreq = Integer.parseInt(frequencySTR);
                        double centerfreq = config.getPreferenceI("AFREQUENCY", 1500);
                        //Limit it's values too
                        if (centerfreq > 2500) centerfreq = 2500;
                        if (centerfreq < 500) centerfreq = 500;
                        String modemInitResult = initCModem(centerfreq);
                        //debugging only
                        //Message.addEntryToLog(Message.dateTimeStamp() + "Done 'initCModem'");
                        //Android Debug
                        //   Modem.MonitorString = modemInitResult;
                        //Prepare RSID Modem
                        createRsidModem();
                        //debugging only
                        //Message.addEntryToLog(Message.dateTimeStamp() + "Done 'createRsidModem'");
                    }
                    //debugging only
                    //Message.addEntryToLog(Message.dateTimeStamp() + "Starting Rx loop");
                    while (RxON) {
                        endproctime = System.currentTimeMillis();
                        double buffertime = (double) numSamples8K / 8000.0 * 1000.0; //in milliseconds
                        if (numSamples8K > 0)
                            Processor.cpuload = (int) (((double) (endproctime - startproctime)) / buffertime * 100);
                        if (Processor.cpuload > 100) Processor.cpuload = 100;
                        Tivar.mHandler.post(Tivar.updatecpuload);
                        //Android try faster mode changes by having a smaller buffer to process						numSamples8K = audiorecorder.read(so8K, 0, 8000/4); //process only part of the buffer to avoid lumpy processing
                        numSamples8K = audiorecorder.read(so8K, 0, 8000 / 8); //process only part of the buffer to avoid lumpy processing
                        if (numSamples8K > 0) {
                            modemState = RXMODEMRUNNING;
                            startproctime = System.currentTimeMillis();
                            //Process only if Rx is ON, otherwise discard (we have already decided to TX)
                            if (RxON) {
                                if (rxRsidOn || (Tivar.currentview == Tivar.MODEMVIEWwithWF)) {
                                    //Re-sample to 11025Hz for RSID, THOR and MFSK modems
                                    int numSamples12K = SampleRateConversion.Process(so8K, numSamples8K, so12K, size12Kbuf);
                                    //Conditional Rx RSID (keep the FFT processing for the waterfall)
                                    rsidReturnedString = RsidCModemReceive(so12K, numSamples12K, rxRsidOn);
                                    //As we flushed the RX pipe automatically, we need to process
                                    // any left-over characters from the previous modem (critical
                                    // for MT63 and Olivia in particular)
                                    //Done in processRxChar() below now
                                    //Modem.MonitorString += rsidReturnedString;
                                    //Processing of the characters received
                                    for (int i = 0; i < rsidReturnedString.length(); i++) {
                                        processRxChar(rsidReturnedString.charAt(i));
                                    }
                                } else {
                                    rsidReturnedString = "";
                                }
                                if (rxRsidOn) {
                                    //if (rsidReturnedString.startsWith("\nRSID:")) {
                                    if (rsidReturnedString.contains("\nRSID:")) {
                                        //We have a new modem and/or centre frequency
                                        //Modem.MonitorString += rsidReturnedString; //Done elsewhere now
                                        //Update the RSID waterfall frequency too
                                        frequency = getCurrentFrequency();
                                        Processor.RxModem = getCurrentMode();
                                        Tivar.saveLastModeUsed(Processor.RxModem);
                                        Tivar.mHandler.post(Tivar.updatetitle);
                                    }
                                }
                                //Sets the latest squelch level for the modem to use
                                setSquelchLevel(squelch);
                                //Flush the text log if no character received for 2 seconds
                                //Safe to flush here since we have not yet processed new audio data yet
                                if (System.currentTimeMillis() > lastLogWriteTime + 2000) {
                                    if (BroadcastLine.length() > 0) {
                                        Broadcastlog.writeBcLog(BroadcastLine.toString());
                                        //Save the time of the last log save
                                        lastLogWriteTime = System.currentTimeMillis();
                                        BroadcastLine = new StringBuilder();
                                    }
                                }
                                //Then process the RX data
                                modemReturnedString = rxCProcess(so8K, numSamples8K);
                                //Retreive latest signal quality for display
                                //Remove averaging metric = misc.decayavg(metric, getMetric(), 20);
                                metric = getMetric();
                                //metric = getMetric();
                                Tivar.mHandler.post(Tivar.updatesignalquality);
                                //Now process all the characters received
                                for (int i = 0; i < modemReturnedString.length(); i++) {
                                    processRxChar(modemReturnedString.charAt(i));
                                }
                            }
                            //Post to monitor (Modem) window after each buffer processing
                            //Add TX frame too if present
                            if (Modem.MonitorString.length() > 0) {
                                Processor.monitor.append(Modem.MonitorString.toString());
                                Modem.MonitorString = new StringBuilder();
                                Tivar.mHandler.post(Tivar.addtomodem);
                            }
                        }
                    }//while (RxON)
                    //We dropped here on pause flag
                    if (audiorecorder != null) {
                        //Avoid some crashes on wrong state
                        if (audiorecorder.getState() == AudioRecord.STATE_INITIALIZED) {
                            if (audiorecorder.getRecordingState() == AudioRecord.RECORDSTATE_RECORDING) {
                                audiorecorder.stop();
                            }
                            audiorecorder.release();
                        }
                    }
                    modemState = RXMODEMPAUSED;
                    //Marker for end of thread (Stop modem thread flag)
                    if (!modemThreadOn) {
                        modemState = RXMODEMIDLE;
                        return;
                    }
                    //Now waits for a restart (or having this thread killed)
                    Processor.restartRxModem.acquireUninterruptibly(1);
                    //Make sure we don's have spare permits
                    Processor.restartRxModem.drainPermits();
                }//while (modemThreadOn)
                //We dropped here on thread stop request
                modemState = RXMODEMIDLE;
            } //run
        }).start(); //	new Thread(new Runnable() {

    }


    public static void stopRxModem() {
        modemThreadOn = false;
        RxON = false;
    }

    public static void pauseRxModem() {
        RxON = false;
    }

    public static void unPauseRxModem() {
        Processor.restartRxModem.release(1);
    }

    static void changemode(int newMode) {
        //Stop the modem receiving side to prevent using the wrong values
        pauseRxModem();
        //Restart modem reception
        unPauseRxModem();

    }

    static void setFrequency(double rxfreq) {
        frequency = rxfreq;
    }

    static void reset() {
        String frequencySTR = config.getPreferenceS("AFREQUENCY", "1500");
        frequency = Integer.parseInt(frequencySTR);
        if (frequency < 500) frequency = 500;
        if (frequency > 2500) frequency = 2500;
        squelch = Tivar.mysp.getFloat("SQUELCHVALUE", (float) 20.0);
    }


    /**
     * @param squelchdiff the delta to add to squelch
     */
    public static void AddtoSquelch(double squelchdiff) {
        squelch += (squelch > 10) ? squelchdiff : squelchdiff / 2;
        if (squelch < 0) squelch = 0;
        if (squelch > 100) squelch = 100;
        //store value into preferences
        SharedPreferences.Editor editor = Tivar.mysp.edit();
        editor.putFloat("SQUELCHVALUE", (float) squelch);
        // Commit the edits!
        editor.commit();
    }

    public static void processRxChar(char inChar) {

        switch (inChar) {
            case 0:
                break; // do nothing
            case 10: //Line Feed
                MonitorString.append("\n");
                BroadcastLine.append("\n");
                break;
            case 13: //ignore Carriage Returns
                break;
            default:
                break;
        }    // end switch

        if (inChar > 31) //Display non-control characters
        {
            MonitorString.append(inChar);
            BroadcastLine.append(inChar);
            if (BroadcastLine.length() > 20) {
                Broadcastlog.writeBcLog(BroadcastLine.toString());
                //Save the time of the last log save
                lastLogWriteTime = System.currentTimeMillis();
                BroadcastLine = new StringBuilder();
            }
        }
    }

}
